""" Python package to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """


# this was inspired by osx-tags by "Ben S / scooby" and is published under
# the same MIT license. See: https://github.com/scooby/osx-tags

import datetime
import json
import logging
import pathlib
import plistlib
import sys

# plistlib creates constants at runtime which causes pylint to complain
from plistlib import FMT_BINARY  # pylint: disable=E0611

import xattr

from ._version import __version__
from .attributes import ATTRIBUTES, Attribute, validate_attribute_value
from .classes import _AttributeList, _AttributeTagsList
from .constants import (
    _COLORIDS,
    _COLORNAMES,
    _FINDER_COMMENT_NAMES,
    _MAX_FINDERCOMMENT,
    _MAX_WHEREFROM,
    _VALID_COLORIDS,
    _kMDItemUserTags,
    kMDItemAuthors,
    kMDItemComment,
    kMDItemCopyright,
    kMDItemCreator,
    kMDItemDescription,
    kMDItemDownloadedDate,
    kMDItemFinderComment,
    kMDItemHeadline,
    kMDItemKeywords,
    kMDItemUserTags,
    kMDItemWhereFroms,
)
from .utils import (
    _debug,
    _get_logger,
    _set_debug,
    clear_finder_comment,
    datetime_naive_to_utc,
    datetime_remove_tz,
    datetime_utc_to_local,
    set_finder_comment,
)

__all__ = [
    "OSXMetaData",
    "__version__",
    "ATTRIBUTES",
    "kMDItemAuthors",
    "kMDItemComment",
    "kMDItemCopyright",
    "kMDItemCreator",
    "kMDItemDescription",
    "kMDItemDownloadedDate",
    "kMDItemFinderComment",
    "kMDItemHeadline",
    "kMDItemKeywords",
    "kMDItemUserTags",
    "_kMDItemUserTags",
    "kMDItemWhereFroms",
]


# TODO: What to do about colors
# TODO: check what happens if OSXMetaData.__init__ called with invalid file--should result in error but saw one case where it didn't
# TODO: cleartags does not always clear colors--this is a new behavior, did Mac OS change something in implementation of colors?


class OSXMetaData:
    """Create an OSXMetaData object to access file metadata"""

    __slots__ = [
        "_fname",
        "_posix_name",
        "_attrs",
        "_tz_aware",
        "__init",
        "authors",
        "comment",
        "copyright",
        "creator",
        "description",
        "downloadeddate",
        "findercomment",
        "headline",
        "keywords",
        "tags",
        "wherefroms",
    ]

    def __init__(self, fname, tz_aware=False):
        """Create an OSXMetaData object to access file metadata
            fname: filename to operate on
            timezone_aware: bool; if True, date/time attributes will return 
                      timezone aware datetime.dateime attributes; if False (default)
                      date/time attributes will return timezone naive objects """
        self._fname = pathlib.Path(fname)
        self._posix_name = self._fname.resolve().as_posix()
        self._tz_aware = tz_aware

        if not self._fname.exists():
            raise FileNotFoundError("file does not exist: ", fname)

        self._attrs = xattr.xattr(self._fname)

        # create property classes for the multi-valued attributes
        # tags get special handling due to color labels
        # ATTRIBUTES contains both long and short names, want only the short names (attribute.name)
        for name in set([attribute.name for attribute in ATTRIBUTES.values()]):
            attribute = ATTRIBUTES[name]
            if attribute.class_ not in [str, float, datetime.datetime]:
                super().__setattr__(
                    name, attribute.class_(attribute, self._attrs, self)
                )

        # Done with initialization
        self.__init = True

    @property
    def name(self):
        """ POSIX path of the file OSXMetaData is operating on """
        return self._fname.resolve().as_posix()

    @property
    def tz_aware(self):
        """ returns the timezone aware flag """
        return self._tz_aware

    @tz_aware.setter
    def tz_aware(self, tz_flag):
        """ sets the timezone aware flag
            tz_flag: bool """
        self._tz_aware = tz_flag

    def _to_dict(self):
        """ Return dict with all attributes for this file 
            key of dict is filename and value is another dict with attributes """

        attribute_list = self.list_metadata()
        json_data = {}
        json_data["_version"] = __version__
        json_data["_filepath"] = self._posix_name
        json_data["_filename"] = self._fname.name
        for attr in attribute_list:
            try:
                attribute = ATTRIBUTES[attr]
                if attribute.type_ == datetime.datetime:
                    # need to convert datetime.datetime to string to serialize
                    value = self.get_attribute(attribute.name)
                    if type(value) == list:
                        value = [v.isoformat() for v in value]
                    else:
                        value = value.isoformat()
                    json_data[attribute.constant] = value
                else:
                    # get raw value
                    json_data[attribute.constant] = self.get_attribute(attribute.name)
            except KeyError:
                # unknown attribute, ignore it
                pass
        return json_data

    def to_json(self):
        """ Returns a string in JSON format for all attributes in this file """
        json_data = self._to_dict()
        json_str = json.dumps(json_data)
        return json_str

    def _restore_attributes(self, attr_dict):
        """ restore attributes from attr_dict
            for each attribute in attr_dict, will set the attribute
            will not clear/erase any attributes on file that are not in attr_dict
            attr_dict: an attribute dict as produced by OSXMetaData._to_dict() """

        for key, val in attr_dict.items():
            if key.startswith("_"):
                # skip private keys like _version and _filepath
                continue
            try:
                self.set_attribute(key, val)
            except:
                logging.warning(f"Unable to restore attribute {attr} for {self._fname}")

    def get_attribute(self, attribute_name):
        """ load attribute and return value or None if attribute was not set 
            attribute_name: name of attribute
        """

        attribute = ATTRIBUTES[attribute_name]
        logging.debug(f"get: {attribute}")

        # user tags need special processing to normalize names
        if attribute.name == "tags":
            self.tags._load_data()
            return self.tags.data

        try:
            plist = plistlib.loads(self._attrs[attribute.constant])
        except KeyError:
            plist = None

        # add UTC to any datetime.datetime objects because that's how MacOS stores them
        # In the plist associated with extended metadata attributes, times are stored as:
        # <date>2020-04-14T14:49:22Z</date>
        if plist and isinstance(plist, list):
            if isinstance(plist[0], datetime.datetime):
                plist = [datetime_naive_to_utc(d) for d in plist]
                if not self._tz_aware:
                    # want datetimes in naive format
                    plist = [
                        datetime_remove_tz(d_local)
                        for d_local in [datetime_utc_to_local(d_utc) for d_utc in plist]
                    ]
        elif isinstance(plist, datetime.datetime):
            plist = datetime_naive_to_utc(plist)
            if not self._tz_aware:
                # want datetimes in naive format
                plist = datetime_remove_tz(datetime_utc_to_local(plist))

        if attribute.as_list and isinstance(plist, list):
            return plist[0]
        else:
            return plist

    def get_attribute_str(self, attribute_name):
        """ returns a string representation of attribute value
            e.g. if attribute is a datedate.datetime object, will 
            format using datetime.isoformat()
            attribute_name: name of attribute """
        value = self.get_attribute(attribute_name)
        if type(value) == list or type(value) == set:
            if type(value[0]) == datetime.datetime:
                new_value = [v.isoformat() for v in value]
                return str(new_value)
            return str(value)
        else:
            if type(value) == datetime.datetime:
                return value.isoformat()
            return value

    def set_attribute(self, attribute_name, value):
        """ write attribute to file with value
            attribute_name: an osxmetadata Attribute name
            value: value to store in attribute """
        attribute = ATTRIBUTES[attribute_name]

        # user tags need special processing to normalize names
        if attribute.name == "tags":
            return self.tags.set_value(value)

        # verify type is correct
        value = validate_attribute_value(attribute, value)

        if attribute.name in _FINDER_COMMENT_NAMES:
            # Finder Comment needs special handling
            # code following will also set the attribute for Finder Comment
            set_finder_comment(self._posix_name, value)
        elif attribute.class_ in [_AttributeList, _AttributeTagsList]:
            getattr(self, attribute.name).set_value(value)
        else:
            # must be a normal scalar (e.g. str, float)
            plist = plistlib.dumps(value, fmt=FMT_BINARY)
            self._attrs.set(attribute.constant, plist)

        return value

    def update_attribute(self, attribute_name, value):
        """ Update attribute with union of itself and value
            (this avoids adding duplicate values to attribute)
            attribute: an osxmetadata Attribute name
            value: value to append to attribute """
        return self.append_attribute(attribute_name, value, update=True)

    def append_attribute(self, attribute_name, value, update=False):
        """ append value to attribute
            attribute_name: an osxmetadata Attribute name
            value: value to append to attribute
            update: (bool) if True, update instead of append (e.g. avoid adding duplicates)
                    (default is False) """

        attribute = ATTRIBUTES[attribute_name]
        logging.debug(f"append_attribute: {attribute} {value}")

        # start with existing values
        new_value = self.get_attribute(attribute.name)

        value = validate_attribute_value(attribute, value)

        if attribute.list:
            if new_value is not None:
                new_value = list(new_value)
                if update:
                    for val in value:
                        if val not in new_value:
                            new_value.append(val)
                else:
                    new_value.extend(value)
            else:
                # no original value
                new_value = value
        else:
            # scalar value
            if update:
                raise AttributeError(f"Cannot use update on {attribute.type_}")
            if new_value is not None:
                new_value += value
            else:
                new_value = value

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                # code following will also set the attribute for Finder Comment
                set_finder_comment(self._posix_name, new_value)
                plist = plistlib.dumps(new_value, fmt=FMT_BINARY)
                self._attrs.set(attribute.constant, plist)
            elif attribute.class_ in [_AttributeList, _AttributeTagsList]:
                # if tags, set_value will normalize
                getattr(self, attribute.name).set_value(new_value)
            else:
                plist = plistlib.dumps(new_value, fmt=FMT_BINARY)
                self._attrs.set(attribute.constant, plist)
        except Exception as e:
            # todo: should catch this or not?
            raise e

        return new_value

    def remove_attribute(self, attribute_name, value):
        """ remove a value from attribute, raise ValueError if attribute does not contain value
            only applies to multi-valued attributes, otherwise raises TypeError
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("remove only applies to multi-valued attributes")

        if attribute.name == "tags":
            # tags need special processing
            self.tags.remove(value)
        else:
            values = self.get_attribute(attribute.name)
            values = list(values)
            values.remove(value)
            self.set_attribute(attribute.name, values)

    def discard_attribute(self, attribute_name, value):
        """ remove a value from attribute, unlike remove, does not raise exception
            if attribute does not contain value
            only applies to multi-valued attributes, otherwise raises TypeError
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("discard only applies to multi-valued attributes")

        values = self.get_attribute(attribute.name)
        try:
            values.remove(value)
            self.set_attribute(attribute.name, values)
        except:
            pass

    def clear_attribute(self, attribute_name):
        """ clear anttribute (remove it from the file) 
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                # code following will also clear the attribute for Finder Comment
                clear_finder_comment(self._posix_name)
            self._attrs.remove(attribute.constant)
        except (IOError, OSError):
            # TODO: fix this try/except handling
            pass

    def _list_attributes(self):
        """ list the attributes set on the file """
        return self._attrs.list()

    def list_metadata(self):
        """ list the Apple metadata attributes set on the file:
            e.g. those in com.apple.metadata namespace """
        # also lists com.osxmetadata.test used for debugging
        mdlist = self._attrs.list()
        mdlist = [
            md
            for md in mdlist
            if md.startswith("com.apple.metadata")
            or md.startswith("com.osxmetadata.test")
        ]
        return mdlist

    def __getattr__(self, name):
        """ if attribute name is in ATTRIBUTE dict, return the value
            otherwise raise KeyError """
        value = self.get_attribute(name)
        return value

    def __setattr__(self, name, value):
        """ if object is initialized and name is an attribute in ATTRIBUTES, 
            set the attribute to value
            if value value is None, will clear (delete) the attribute and all associated values
            if name is not a metadata attribute, assume it's a normal class attribute and pass to
            super() to handle  """
        try:
            if self.__init:
                # already initialized
                attribute = ATTRIBUTES[name]
                if value is None:
                    self.clear_attribute(attribute.name)
                else:
                    self.set_attribute(attribute.name, value)
        except (KeyError, AttributeError):
            super().__setattr__(name, value)



import datetime
import json
import logging
import os

from . import _applescript

_DEBUG = False

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(message)s",
)

if not _DEBUG:
    logging.disable(logging.DEBUG)


def _get_logger():
    """Used only for testing
    
    Returns:
        logging.Logger object -- logging.Logger object for osxmetadata
    """
    return logging.Logger(__name__)


def _set_debug(debug):
    """ Enable or disable debug logging """
    global _DEBUG
    _DEBUG = debug
    if debug:
        logging.disable(logging.NOTSET)
    else:
        logging.disable(logging.DEBUG)


def _debug():
    """ returns True if debugging turned on (via _set_debug), otherwise, false """
    return _DEBUG


_scpt_set_finder_comment = _applescript.AppleScript(
    """
            on run {fname, fc}
	            set theFile to fname
	            set theComment to fc
	            tell application "Finder" to set comment of (POSIX file theFile as alias) to theComment
            end run
            """
)

_scpt_clear_finder_comment = _applescript.AppleScript(
    """
            on run {fname}
	            set theFile to fname
	            set theComment to missing value
	            tell application "Finder" to set comment of (POSIX file theFile as alias) to theComment
            end run
            """
)


def set_finder_comment(filename, comment):
    """ set finder comment for filename
        filename: path to file in posix format
        comment: comment string
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"Could not find {filename}")

    _scpt_set_finder_comment.run(filename, comment)


def clear_finder_comment(filename):
    """ clear finder comment for filename
        filename: path to file in posix format
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"Could not find {filename}")

    _scpt_clear_finder_comment.run(filename)


def load_backup_file(backup_file):
    """ Load attribute data from JSON in backup_file 
        Returns: backup_data dict """

    if not os.path.isfile(backup_file):
        raise FileNotFoundError(f"Could not find backup file: {backup_file}")

    backup_data = {}
    fp = open(backup_file, mode="r")

    for line in fp:
        data = json.loads(line)
        fname = data["_filename"]
        if fname in backup_data:
            logging.warning(
                f"WARNING: duplicate filename {fname} found in {backup_file}"
            )

        backup_data[fname] = data

    fp.close()
    return backup_data


def write_backup_file(backup_file, backup_data):
    """ Write backup_data to backup_file as JSON
        backup_data: dict where key is filename and value is dict of the attributes
        as returned by json.loads(OSXMetaData.to_json()) """

    fp = open(backup_file, mode="w")

    for record in backup_data.values():
        json_str = json.dumps(record)
        fp.write(json_str)
        fp.write("\n")

    fp.close()


# datetime.datetime helper functions for converting to/from UTC


def get_local_tz():
    """ return local timezone as datetime.timezone tzinfo """
    local_tz = (
        datetime.datetime.now(datetime.timezone(datetime.timedelta(0)))
        .astimezone()
        .tzinfo
    )
    return local_tz


def datetime_has_tz(dt):
    """ return True if datetime dt has tzinfo else False
        dt: datetime.datetime
        returns True if dt is timezone aware, else False """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        return True
    return False


def datetime_tz_to_utc(dt):
    """ convert datetime.datetime object with timezone to UTC timezone 
        dt: datetime.datetime object
        returns: datetime.datetime in UTC timezone """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    local_tz = get_local_tz()
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        dt_utc = dt.replace(tzinfo=dt.tzinfo).astimezone(tz=datetime.timezone.utc)
        return dt_utc
    else:
        raise ValueError(f"dt does not have timezone info")


def datetime_remove_tz(dt):
    """ remove timezone from a datetime.datetime object
        dt: datetime.datetime object with tzinfo
        returns: dt without any timezone info (naive datetime object) """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    dt_new = dt.replace(tzinfo=None)
    return dt_new


def datetime_naive_to_utc(dt):
    """ convert naive (timezone unaware) datetime.datetime
        to aware timezone in UTC timezone
        dt: datetime.datetime without timezone
        returns: datetime.datetime with UTC timezone """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        # has timezone info
        raise ValueError(
            "dt must be naive/timezone unaware: "
            f"{dt} has tzinfo {dt.tzinfo} and offset {dt.tzinfo.utcoffset(dt)}"
        )

    dt_utc = dt.replace(tzinfo=datetime.timezone.utc)
    return dt_utc


def datetime_naive_to_local(dt):
    """ convert naive (timezone unaware) datetime.datetime
        to aware timezone in local timezone
        dt: datetime.datetime without timezone
        returns: datetime.datetime with local timezone """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        # has timezone info
        raise ValueError(
            "dt must be naive/timezone unaware: "
            f"{dt} has tzinfo {dt.tzinfo} and offset {dt.tizinfo.utcoffset(dt)}"
        )

    dt_local = dt.replace(tzinfo=get_local_tz())
    return dt_local


def datetime_utc_to_local(dt):
    """ convert datetime.datetime object in UTC timezone to local timezone 
        dt: datetime.datetime object
        returns: datetime.datetime in local timezone """

    if type(dt) != datetime.datetime:
        raise TypeError(f"dt must be type datetime.datetime, not {type(dt)}")

    if dt.tzinfo is not datetime.timezone.utc:
        raise ValueError(f"{dt} must be in UTC timezone: timezone = {dt.tzinfo}")

    dt_local = dt.replace(tzinfo=datetime.timezone.utc).astimezone(tz=get_local_tz())
    return dt_local



""" Classes for metadata attribute types """

import collections.abc
import datetime
import plistlib
from plistlib import FMT_BINARY  # pylint: disable=E0611
import sys

from .constants import _COLORNAMES, _VALID_COLORIDS
from .utils import datetime_naive_to_utc, datetime_utc_to_local, datetime_remove_tz


class _AttributeList(collections.abc.MutableSequence):
    """ represents a multi-valued OSXMetaData attribute list """

    def __init__(self, attribute, xattr_, osxmetadata_obj):
        """ initialize object
            attribute: an OSXMetaData Attributes namedtuple 
            xattr_: an instance of xattr.xattr
            osxmetadata_obj: instance of OSXMetaData that created this class instance """
        self._attribute = attribute
        self._attrs = xattr_
        self._md = osxmetadata_obj

        self._constant = attribute.constant

        self.data = []
        self._values = []
        self._load_data()

    def set_value(self, value):
        self.data = value
        self._write_data()

    def _load_data(self):
        self._values = []
        try:
            # load the binary plist value
            self._values = plistlib.loads(self._attrs[self._constant])
            if self._values:
                try:
                    self.data = list(self._values)
                    if self._attribute.type_ == datetime.datetime:
                        # add UTC timezone to datetime values
                        # because that's how MacOS stores them
                        # In the plist associated with xattr, times are stored as:
                        # <date>2020-04-14T14:49:22Z</date>
                        self.data = [datetime_naive_to_utc(dt) for dt in self.data]
                        if not self._md._tz_aware:
                            # want datetimes in naive format
                            self.data = [
                                datetime_remove_tz(d_local)
                                for d_local in [
                                    datetime_utc_to_local(d_utc) for d_utc in self.data
                                ]
                            ]
                except TypeError:
                    self.data = set([self._values])
                    if self._attribute.type_ == datetime.datetime:
                        # add UTC timezone to datetime values
                        self.data = {datetime_naive_to_utc(dt) for dt in self.data}
                        if not self._tz_aware:
                            # want datetimes in naive format
                            self.data = {
                                datetime_remove_tz(d_local)
                                for d_local in {
                                    datetime_utc_to_local(d_utc) for d_utc in self.data
                                }
                            }
            else:
                self.data = []
        except KeyError:
            self.data = []

    def sort(self, key=None, reverse=False):
        self._load_data()
        self.data.sort(key=key, reverse=reverse)
        self._write_data()

    def __delitem__(self, index):
        self._load_data()
        self.data.__delitem__(index)
        self._write_data()

    def __getitem__(self, index):
        self._load_data()
        return self.data.__getitem__(index)

    def __len__(self):
        self._load_data()
        return self.data.__len__()

    def __setitem__(self, index, value):
        self._load_data()
        self.data.__setitem__(index, value)
        self._write_data()
        self._load_data()

    def insert(self, index, value):
        self._load_data()
        self.data.insert(index, value)
        self._write_data()

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        plist = plistlib.dumps(self.data, fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)

    def __repr__(self):
        # return f"{type(self).__name__}({super().__repr__()})"
        self._load_data()
        return repr(self.data)

    def __eq__(self, other):
        self._load_data()
        return self.data == other


class _AttributeTagsList(_AttributeList):
    """ represents a _kMDItemUserTag attribute list """

    def _tag_split(self, tag):
        # Extracts the color information from a Finder tag.

        parts = tag.rsplit("\n", 1)
        if len(parts) == 1:
            return parts[0], 0
        elif (
            len(parts[1]) != 1 or parts[1] not in _VALID_COLORIDS
        ):  # Not a color number
            return tag, 0
        else:
            return parts[0], int(parts[1])

    def _normalize(self, tag):
        """
        Ensures a color is set if not none.
        :param tag: a possibly non-normal tag.
        :return: A colorized tag.
        """
        tag, color = self._tag_split(tag)
        if tag.title() in _COLORNAMES:
            # ignore the color passed and set proper color name
            return self._tag_colored(tag.title(), _COLORNAMES[tag.title()])
        else:
            return self._tag_colored(tag, color)

    def _tag_nocolor(self, tag):
        """
        Removes the color information from a Finder tag.
        """
        return tag.rsplit("\n", 1)[0]

    def _tag_colored(self, tag, color):
        """
        Sets the color of a tag.

        Parameters:
        tag(str): a tag name
        color(int): an integer from 1 through 7

        Return:
        (str) the tag with encoded color.
        """
        return f"{self._tag_nocolor(tag)}\n{color}"

    def _load_data(self):
        self._tags = {}
        try:
            self._tagvalues = self._attrs[self._constant]
            # load the binary plist value
            self._tagvalues = plistlib.loads(self._tagvalues)
            for x in self._tagvalues:
                (tag, color) = self._tag_split(x)
                self._tags[tag] = color
        except KeyError:
            self._tags = None
        if self._tags:
            # ZZZ --> keys drops duplicate tags
            # self.data = list(self._tags.keys())
            # separate the tag from the color id
            self.data = [self._tag_split(value)[0] for value in self._tagvalues]
        else:
            self.data = []

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        self._tagvalues = list(map(self._normalize, self.data))
        plist = plistlib.dumps(self._tagvalues, fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)


# class _AttributeSet:
#     """ represents a multi-valued OSXMetaData attribute set """

#     def __init__(self, attribute, xattr_, osxmetadata_obj):
#         """ initialize object
#             attribute: an OSXMetaData Attributes namedtuple
#             xattr_: an instance of xattr.xattr """
#             osxmetadata_obj: instance of OSXMetaData that created this class instance """
#         self._attribute = attribute
#         self._attrs = xattr_
#         self._md = osxmetadata_obj
#         self._constant = attribute.constant

#         # initialize
#         self.data = set()
#         self._load_data()

#     def set_value(self, values):
#         """ set value to values """
#         self.data = set(map(self._normalize, values))
#         self._write_data()

#     def add(self, value):
#         """ add a value"""
#         # TODO: should check to see if value is a non-list, set, etc. (single value)
#         self._load_data()
#         self.data.add(self._normalize(value))
#         self._write_data()

#     def update(self, *others):
#         """ update data adding any new values in *others
#             each item passed in *others must be an iterable """
#         self._load_data()
#         old_values = set(map(self._normalize, self.data))
#         new_values = old_values
#         for item in others:
#             new_values = new_values.union(set(map(self._normalize, item)))
#         self.data = new_values
#         self._write_data()

#     def clear(self):
#         """ clear attribute (removes all values) """
#         try:
#             self._attrs.remove(self._constant)
#         except (IOError, OSError):
#             pass

#     def remove(self, value):
#         """ remove a value, raise ValueError exception if value does not exist in data set """
#         self._load_data()
#         if value not in self.data:
#             raise ValueError("list.remove(x): x not in list")
#         values = set(map(self._normalize, self.data))
#         values.remove(self._normalize(value))
#         self.data = values
#         self._write_data()

#     def discard(self, value):
#         """ remove a value, does not raise exception if value does not exist """
#         self._load_data()
#         values = set(map(self._normalize, self.data))
#         values.discard(self._normalize(value))
#         self.data = values
#         self._write_data()

#     def _load_data(self):
#         self._values = []
#         try:
#             # load the binary plist value
#             self._values = plistlib.loads(self._attrs[self._constant])
#             if self._values:
#                 try:
#                     self.data = set(self._values)
#                 except TypeError:
#                     self.data = set([self._values])
#             else:
#                 self.data = set()
#         except KeyError:
#             self.data = set()

#     def _write_data(self):
#         # Overwrites the existing tags with the iterable of tags provided.
#         plist = plistlib.dumps(list(map(self._normalize, self.data)), fmt=FMT_BINARY)
#         self._attrs.set(self._constant, plist)

#     def _normalize(self, value):
#         """ processes a value to normalize/transform the value if needed
#             override in sublcass if desired (e.g. used _TagsSet) """
#         return value

#     def __iter__(self):
#         self._load_data()
#         for value in self.data:
#             yield value

#     def __len__(self):
#         self._load_data()
#         return len(self.data)

#     def __repr__(self):
#         self._load_data()
#         return repr(self.data)

#     def __str__(self):
#         self._load_data()
#         if self._attribute.type_ == datetime.datetime:
#             values = [d.isoformat() for d in self.data]
#         else:
#             values = self.data
#         return str(list(values))

#     def __ior__(self, values):
#         if type(values) != set:
#             raise TypeError
#         self.update(values)
#         return self


# deprecated
# class _AttributeTagsSet(_AttributeSet):
#     """ represents a _kMDItemUserTag attribute set """

#     def _tag_split(self, tag):
#         # Extracts the color information from a Finder tag.

#         parts = tag.rsplit("\n", 1)
#         if len(parts) == 1:
#             return parts[0], 0
#         elif (
#             len(parts[1]) != 1 or parts[1] not in _VALID_COLORIDS
#         ):  # Not a color number
#             return tag, 0
#         else:
#             return parts[0], int(parts[1])

#     def _normalize(self, tag):
#         """
#         Ensures a color is set if not none.
#         :param tag: a possibly non-normal tag.
#         :return: A colorized tag.
#         """
#         tag, color = self._tag_split(tag)
#         if tag.title() in _COLORNAMES:
#             # ignore the color passed and set proper color name
#             return self._tag_colored(tag.title(), _COLORNAMES[tag.title()])
#         else:
#             return self._tag_colored(tag, color)

#     def _tag_nocolor(self, tag):
#         """
#         Removes the color information from a Finder tag.
#         """
#         return tag.rsplit("\n", 1)[0]

#     def _tag_colored(self, tag, color):
#         """
#         Sets the color of a tag.

#         Parameters:
#         tag(str): a tag name
#         color(int): an integer from 1 through 7

#         Return:
#         (str) the tag with encoded color.
#         """
#         return f"{self._tag_nocolor(tag)}\n{color}"

#     def _load_data(self):
#         self._tags = {}
#         try:
#             self._tagvalues = self._attrs[self._constant]
#             # load the binary plist value
#             self._tagvalues = plistlib.loads(self._tagvalues)
#             for x in self._tagvalues:
#                 (tag, color) = self._tag_split(x)
#                 self._tags[tag] = color
#         except KeyError:
#             self._tags = None
#         if self._tags:
#             self.data = set(self._tags.keys())
#         else:
#             self.data = set()
